@@ -1,4 +1,5 @@
 import base64
+import glob
 import os
 import pathlib
 import re
@@ -11,11 +12,11 @@ import numpy as np
 from appium.webdriver.common.touch_action import TouchAction
 from appium.webdriver.webdriver import WebDriver
 from selenium.common.exceptions import TimeoutException
-from typing import List
+from typing import List, Iterator
 
-VERTICAL_ORDER = 0
-HORIZONTAL_ORDER = 1
-SIMILARITY_ORDER = 2
+SIMILARITY_ORDER = 0
+VERTICAL_ORDER = 1
+HORIZONTAL_ORDER = 2
 
 CV_CCOEFF = cv.TM_CCOEFF_NORMED
 CV_CCORR = cv.TM_CCORR_NORMED
@@ -24,18 +25,22 @@ CV_SQDIFF = cv.TM_SQDIFF_NORMED
 
 class VisualMatch(object):
 
-    def __init__(self, x, y, width, height, similarity):
+    def __init__(self, driver, x, y, width, height, similarity):
+        self.driver = driver
         self.x = x
         self.y = y
         self.width = width
         self.height = height
         self.similarity = similarity
 
+    def click(self):
+        TouchAction(self.driver).press(x=self.x, y=self.y).release().perform()
+
 
 class VisualElement(object):
 
-    def __init__(self, driver: WebDriver, path, order: int = 0, disposal: int = VERTICAL_ORDER,
-                 similarity: float = 0.8, method=CV_CCOEFF, name: str = None):
+    def __init__(self, driver: WebDriver, path, order: int = 0, disposal: int = SIMILARITY_ORDER,
+                 similarity: float = 0.7, method=CV_CCOEFF, name: str = None):
         self.driver = driver
         self.path = str(path)
         self.order = order
@@ -46,34 +51,36 @@ class VisualElement(object):
 
         self.max_occurrences = 16
         self._debug = False
+        self._searched = False
         self.matches: List[VisualMatch] = []
 
     @property
     def match(self) -> VisualMatch:
-        return self.matches[0] if self.matches else None
+        return self[0]
 
-    def __len__(self):
-        return len(self.matches if self.matches else self.reset_object().matches)
+    def __len__(self) -> int:
+        return len(self.matches if self._searched else self._search_all_matches().matches)
 
-    def __getitem__(self, i):
+    def __getitem__(self, i) -> VisualMatch:
         try:
-            return (self.matches if self.matches else self.reset_object().matches)[i]
+            return (self.matches if self._searched else self._search_all_matches().matches)[i]
         except IndexError:
             raise Exception(f"Index above matches: {self.order} > {len(self.matches)}")
 
-    def __iter__(self):
-        return iter(self.matches if self.matches else self.reset_object().matches)
+    def __iter__(self) -> Iterator[VisualMatch]:
+        return iter(self.matches if self._searched else self._search_all_matches().matches)
 
-    def debug(self):
-        self._debug = True
+    def debug(self, enable=True):
+        self._debug = enable
         return self
 
     def reset_object(self):
-        self._find_matches()
+        self._searched = False
+        self.matches.clear()
         return self
 
     def is_visible(self):
-        return bool(self.reset_object().matches)
+        return bool(self._search_all_matches().matches)
 
     def wait_until_visible(self, timeout=10):
         start_time = datetime.now()
@@ -94,10 +101,10 @@ class VisualElement(object):
             if time_delta.total_seconds() >= timeout:
                 raise TimeoutException(f'Still present "{self.name}" in {timeout} seconds')
             time.sleep(1)
+        return self
 
     def click(self):
-        match = (self.matches if self.matches else self.reset_object().matches)[self.order]
-        TouchAction(self.driver).press(x=match.x, y=match.y).release().perform()
+        self[self.order].click()
         return self
 
     @staticmethod
@@ -106,16 +113,32 @@ class VisualElement(object):
         end_x, end_y = (int((loc[0] + w) * r), int((loc[1] + h) * r))
         return start_x, start_y, end_x, end_y
 
-    def _find_matches(self):
-        if not os.path.isfile(self.path):
-            raise Exception(f"Template route doesn't exist: {self.path}")
+    def _search_all_matches(self):
+        self.reset_object()
+
+        for template_path in glob.glob(self.path):
+            self._find_matches(template_path)
+
+        if self.disposal == VERTICAL_ORDER:
+            self.matches.sort(key=lambda v: v.y)
+        elif self.disposal == HORIZONTAL_ORDER:
+            self.matches.sort(key=lambda v: v.x)
+        elif self.disposal == SIMILARITY_ORDER:
+            self.matches.sort(key=lambda v: v.similarity, reverse=True)  # more similar first
+
+        self._searched = True
+        return self
+
+    def _find_matches(self, template_path):
+        if not os.path.isfile(template_path):
+            raise Exception(f"Template route doesn't exist: {template_path}")
         if self.disposal not in range(2):
             raise Exception(f"Disposal priority not valid: {self.disposal}")
         if self.order not in range(self.max_occurrences):
             raise Exception(f"Order above max occurrences: {self.order} > {self.max_occurrences}")
 
         # load the template image, convert it to grayscale
-        template = cv.imread(self.path)
+        template = cv.imread(template_path)
         template = cv.cvtColor(template, cv.COLOR_BGR2GRAY)
         h, w = template.shape[:2]
 
@@ -126,7 +149,7 @@ class VisualElement(object):
 
         if self._debug:
             os.makedirs('temp', exist_ok=True)
-            cv.imwrite('temp/velenium__screenshot.png', image)
+            cv.imwrite(f'temp/velenium_{int(time.time() * 10**6)}__{self.name}__screenshot.png', image)
 
         gray = cv.cvtColor(image, cv.COLOR_BGR2GRAY)
         occurrence = None
@@ -158,10 +181,9 @@ class VisualElement(object):
             if self._debug:
                 clone = np.dstack([resized, resized, resized])
                 cv.rectangle(clone, (max_loc[0], max_loc[1]), (max_loc[0] + w, max_loc[1] + h), (0, 0, 255), 2)
-                cv.imwrite(f'temp/velenium__{self.name}__{int(max_val * 100)}.png', clone)
+                cv.imwrite(f'temp/velenium_{int(time.time() * 10**6)}__{self.name}__{int(max_val * 100)}.png', clone)
 
         if not occurrence:
-            self.matches = []
             return
 
         # unpack the bookkeeping variable and compute the coordinates of the bounding box based on the resized ratio
@@ -171,18 +193,18 @@ class VisualElement(object):
         # draw a bounding box around the detected result and display the image
         if self._debug:
             cv.rectangle(image, (start_x, start_y), (end_x, end_y), (0, 0, 255), 2)
-            cv.imwrite(f'temp/velenium__{self.name}__best.png', image)
+            cv.imwrite(f'temp/velenium_{int(time.time() * 10**6)}__{self.name}__best.png', image)
 
-        self.matches = [VisualMatch(start_x + int(w / 2), start_y + int(h / 2), w, h, max_val)]
+        self.matches += [VisualMatch(self.driver, start_x + int(w / 2), start_y + int(h / 2), w, h, max_val)]
 
         # from that resized image get
         for i in range(self.max_occurrences):
 
-            # cover the previous best match in black
+            # cover the previous best match in black to discard duplicates
             cv.rectangle(resized, (max_loc[0], max_loc[1]), (max_loc[0] + w, max_loc[1] + h), (0, 0, 0), -1)
 
             if self._debug:
-                cv.imwrite(f'temp/velenium__{self.name}__covered_{i}.png', resized)
+                cv.imwrite(f'temp/velenium_{int(time.time() * 10**6)}__{self.name}__covered_{i}.png', resized)
 
             # repeat template matching to find the template in the image
             result = cv.matchTemplate(resized, template, self.method)
@@ -191,16 +213,9 @@ class VisualElement(object):
                 break
 
             start_x, start_y, end_x, end_y = self._get_bounds(max_loc, w, h, r)
-            self.matches.append(VisualMatch(start_x + int(w / 2), start_y + int(h / 2), w, h, max_val))
+            self.matches.append(VisualMatch(self.driver, start_x + int(w / 2), start_y + int(h / 2), w, h, max_val))
             if self._debug:
                 cv.rectangle(image, (start_x, start_y), (end_x, end_y), (0, 0, 255), 2)
 
         if self._debug:
-            cv.imwrite(f'temp/velenium__{self.name}__all.png', image)
-
-        if self.disposal == VERTICAL_ORDER:
-            self.matches.sort(key=lambda v: v.y)
-        elif self.disposal == HORIZONTAL_ORDER:
-            self.matches.sort(key=lambda v: v.x)
-        elif self.disposal == SIMILARITY_ORDER:
-            self.matches.sort(key=lambda v: v.similarity, reverse=True)  # more similar first
+            cv.imwrite(f'temp/velenium_{int(time.time() * 10**6)}__{self.name}__all.png', image)
